unit GA; {(asexual) Genetic Algorithm}
{Copyright (C) 1994, Jon Shemitz <jon@armory.com> }

interface

uses BigString;

{0: macros}

  function imul_idiv(A, B, C: integer): integer; {A imul B idiv C}
  Inline(
        $59 /			{pop	cx}
        $5B /			{pop	bx}
        $58 /			{pop	ax}
        $F7 / $EB /		{imul	bx}
        $F7 / $F9		{idiv	cx}
        );

  function mul_div(A, B, C: word): word; {A mul B div C}
  Inline(
        $59 /			{pop	cx}
        $5B /			{pop	bx}
        $58 /			{pop	ax}
        $F7 / $E3 /		{mul	bx}
        $F7 / $F1		{div	cx}
        );

{1: Genes}

type
  GenePtr = ^ Genes;
  Genes = object (BigStrings) {a specialized sort of index strings}
          function Mutate: GenePtr;
          function Randomize: GenePtr;
            {Both above fns return modified COPIES; original unchanged}
          procedure Show(var Txt: text; Tag: string);
          end;

{2: Domain specific GA helper fns}

type
  ProblemDescriptionRecord = record
                             First, Subsequent: word; {clusters/disk}
                             Optimal:           word; {# floppies}
                             end;

procedure SetPDR( var PDR: ProblemDescriptionRecord;
                  First, Subsequent, Optimal: word);

function Floppies( const PDR: ProblemDescriptionRecord;
                  Sizes: BigStringPtr; Ordering: GenePtr): word;

function Score( const PDR: ProblemDescriptionRecord;
                Sizes: BigStringPtr; Ordering: GenePtr): LongInt;
{Mean-Square of space occupied (in thousandths of a floppy)
 on all but the last floppy}

const PerfectScore = 1000 * 1000;

function Partition( const PDR: ProblemDescriptionRecord;
                    Sizes: BigStringPtr; Ordering: GenePtr): BigStringPtr;

{ 3: The GA itself }

function Optimize( const PDR:           ProblemDescriptionRecord;
                         Sizes:         BigStringPtr;
                         BreedFrom,
                         BreedTo,
                         MaxIterations: integer;
                   var   Output:        text;
                         Report:        boolean ): GenePtr;

implementation

uses Consts;

{1: Genes}

procedure Genes.Show(var Txt: text; Tag: string);
var
  Idx: word;
begin
  WriteLn(Txt, Tag);
  with Data do
    for Idx := 1 to Wrd[0] do Write(Txt, Wrd[Idx], ' ');
  WriteLn(Txt, ^J);
end;

(*
const
  MultipleMutationRate = 150; {Per thousand}

  ShiftRatio     = 50;
  TransposeRatio = 25;
  ReverseRatio   = 05;

  All_Ratios     = ShiftRatio + TransposeRatio + ReverseRatio;

  ExpoRandRatio = 666; {per thousand}
*)

  function ExpoRand(Max: word): word; near; {1..Max}
  var
    Result: word;
  begin
    Result := 1;
    while (Result < Max) and (Random(1000) + 1 < ExpoRandRatio) do Inc(Result);
    ExpoRand := Result;
  end;

  function Max(A, B: word): word; near;
  begin
    if A > B then Max := A else Max := B;
  end;

  function Min(A, B: word): word; near;
  begin
    if A < B then Min := A else Min := B;
  end;


function Genes.Mutate: GenePtr;
{*
  Performs 1 or more of 3 types of mutation:
  1: Simple shift:   Move 1 or more indices 1 or more spaces fwd or backwards.
  2: Transposition:  Xchg 1 or more indices with same len string,
                     elsewhere in vector
  3: Reversal:       Flip substring, front to back
*}
type
  WordArray = array[1..32000] of word;
var
  Result: GenePtr;
  GeneLength, Length: word;
  Start, Destination: integer;
  Clipboard: ^ WordArray;
begin
  BigStringPtr(Result) := Self.Copy;
  GeneLength := Data.Wrd[0];
  GetMem(Clipboard, GeneLength * SizeOf(Word));
  repeat {single mutation}
    case Random(All_Ratios) of
      0..ShiftRatio-1:
           begin {simple shift}
           Start  := Random(GeneLength-1)+1; {Start anywhere but last pos.}
           Length := ExpoRand(GeneLength - Start);
           repeat
             Destination := Random(GeneLength - Length) + 1;
           until Destination <> Start;
           with Result^.Data do
             begin
             Move(Wrd[Start], Clipboard^, Length * SizeOf(Word)); {Copy out}
             Move( Wrd[Start+Length], Wrd[Start],
                   (GeneLength - (Start + Length - 1)) * SizeOf(Word));
               {Shift tail down}
             Move( Wrd[Destination], Wrd[Destination+Length],
                  ((GeneLength - Length) - (Destination-1)) * SizeOf(Word) );
               {Make room}
             Move(Clipboard^, Wrd[Destination], Length * SizeOf(Word)); {Copy back}
             end;
           end;
      ShiftRatio..ShiftRatio+TransposeRatio-1:
           begin {transposition}
           Start  := Random(GeneLength) + 1;
           repeat
             Destination := Random(GeneLength) + 1;
           until Destination <> Start;
           Length := ExpoRand(
                       Min( Abs(Start - Destination), {SpacesThatCanBeXchged}
                            GeneLength - Max(Start, Destination) + 1) )
                     * SizeOf(word);
           {Up to Min of SpacesThatCanBeXchged or SpaceBetweenLastAndEOL}
           with Result^.Data do
             begin
             Move(Wrd[Start], ClipBoard^, Length);
             Move(Wrd[Destination], Wrd[Start], Length);
             Move(Clipboard^, Wrd[Destination], Length);
             end;
           end;
      else begin {Reversal}
           Start  := Random(GeneLength-1) + 1; {Start anywhere but last pos}
           Length := ExpoRand(GeneLength - Start) + 1; {Can't swap 1 word!}
           with Result^.Data do
             begin
             Move(Wrd[Start], Clipboard^, Length * SizeOf(word));
             for Destination := 0 to Length - 1 do
               Wrd[Start+Destination] := Clipboard^[Length-Destination];
             end;
           end;
      end;
  until Random(1000) + 1 > MultipleMutationRate;
  FreeMem(Clipboard, GeneLength * SizeOf(Word));
  Mutate := Result;
end;

function Genes.Randomize: GenePtr;
var
  Length, Idx,
  Swap, Tmp:   word;
  Result:      BigStringPtr;
begin
  Length := GetLength;
  Result := GenerateIndexVector(Length);
  with Result^ do
    for Idx := Length downto 2 do
      begin
      Swap := Random(Idx) + 1;
      if Swap <> Idx then
        begin
        Tmp := Data.Wrd[Idx];
        Data.Wrd[Idx] := Data.Wrd[Swap];
        Data.Wrd[Swap] := Tmp;
        end;
      end;
  Randomize := GenePtr(Result);
end;

{2: Domain specific GA helper fns}

procedure SetPDR( var PDR: ProblemDescriptionRecord;
                  First, Subsequent, Optimal: word);
begin
  PDR.First      := First;
  PDR.Subsequent := Subsequent;
  PDR.Optimal    := Optimal;
end;

function Floppies( const PDR: ProblemDescriptionRecord;
                   Sizes: BigStringPtr; Ordering: GenePtr): word;
var
  Files, Result, Idx, OnThis, Limit, ThisSize: word;
begin
  with Ordering^ do
    begin
    Files := Data.Wrd[0];
    {$ifopt R+}
      if (StringType <> _Wrd) or (Files = 0) then RunError(201);
      for Idx := 1 to Files do
        if Sizes^.Data.Wrd[ Data.Wrd[Idx] ] > PDR.Subsequent then
          RunError(201); {This file is too big to fit on a floppy}
    {$endif}
    Result := 1;
    Idx    := 1;
    OnThis := 0;
    Limit  := PDR.First;
    repeat
      ThisSize := Sizes^.Data.Wrd[Data.Wrd[Idx]];
      Inc(OnThis, ThisSize);
      if OnThis > Limit then
        begin
        Inc(Result);
        OnThis := ThisSize;
        Limit  := PDR.Subsequent;
        end;
      Inc(Idx);
    until Idx > Files;
    end;
  Floppies := Result;
end;

function Score( const PDR: ProblemDescriptionRecord;
                Sizes: BigStringPtr; Ordering: GenePtr): LongInt;
{Mean-Square of space occupied (in thousandths of a floppy)
 on all but the last floppy}
var
  SumSquares: LongInt;
  Files, Floppies, Idx, OnThis, Limit, ThisSize: word;
begin
  with Ordering^ do
    begin
    Files := Data.Wrd[0];
    {$ifopt R+}
      if (StringType <> _Wrd) or (Files = 0) then RunError(201);
      for Idx := 1 to Files do
        if Sizes^.Data.Wrd[ Data.Wrd[Idx] ] > PDR.Subsequent then
          RunError(201); {This file is too big to fit on a floppy}
    {$endif}
    Floppies := 0;
    Idx      := 1;
    OnThis   := 0;
    Limit    := PDR.First;
    SumSquares := 0;
    repeat
      ThisSize := Sizes^.Data.Wrd[Data.Wrd[Idx]];
      Inc(OnThis, ThisSize);
      if OnThis > Limit then
        begin
        Inc( SumSquares,
             Sqr(LongInt( mul_div( OnThis - ThisSize, 1000, Limit))));
        Inc(Floppies);
        OnThis := ThisSize;
        Limit  := PDR.Subsequent;
        end;
      Inc(Idx);
    until Idx > Files;
    end;
  if Floppies = PDR.Optimal - 1
    then Score := PerfectScore
    else Score := SumSquares div Floppies;
         {Avg square of (full) floppy utilization, in tenth%s}
end;


function Partition( const PDR: ProblemDescriptionRecord;
                    Sizes: BigStringPtr; Ordering: GenePtr): BigStringPtr;
var
  Result: BigStringPtr;
  Files, Idx, EndPt, OnThis, Limit, ThisSize: word;
begin
  Result := New_BigString(_Wrd, Floppies(PDR, Sizes, Ordering) );
  {Last file in ordering on each (full) disk;
   NFiles in last pos ...}
  with Ordering^ do
    begin
    Files := Data.Wrd[0];
    {$ifopt R+}
      if (StringType <> _Wrd) or (Files = 0) then RunError(201);
      for Idx := 1 to Files do
        if Sizes^.Data.Wrd[ Data.Wrd[Idx] ] > PDR.Subsequent then
          RunError(201); {This file is too big to fit on a floppy}
    {$endif}
    Idx    := 1;
    EndPt  := 1;
    OnThis := 0;
    Limit  := PDR.First;
    repeat
      ThisSize := Sizes^.Data.Wrd[Data.Wrd[Idx]];
      Inc(OnThis, ThisSize);
      if OnThis > Limit then
        begin
        Result^.Data.Wrd[EndPt] := Idx  - 1;
        Inc(EndPt);
        OnThis := ThisSize;
        Limit  := PDR.Subsequent;
        end;
      Inc(Idx);
    until Idx > Files;
    Result^.Data.Wrd[EndPt] := Idx  - 1;
    end;
  Partition := Result;
end;

{ 3: The GA itself }

function Optimize( const PDR:           ProblemDescriptionRecord;
                         Sizes:         BigStringPtr;
                         BreedFrom,
                         BreedTo,
                         MaxIterations: integer;
                   var   Output:        text;
                         Report:        boolean ): GenePtr;
var
  Idx, GeneNumber,
  Breeders, GeneIndex,
  Generations: integer;
  ThisGene, BestGene:  GenePtr;
  BestScore, MeanScore:LongInt;
  Duplicate,
  AWinner, TimeOut:    boolean;
  GenePool, Scores, SortVector, BreedingPool: BigStringPtr;

begin
  {Create an initial gene pool}
    GenePool     := New_BigString(_Ptr, BreedTo);
    BreedingPool := New_BigString(_Ptr, BreedFrom);
    ThisGene := GenePtr( GenerateIndexVector(Sizes^.GetLength) ); {1,2,3,...,N}
    GenePool^.Data.Ptr[1] := ThisGene; {Why not use it as a candidate gene?}
    for Idx := 2 to BreedTo do
      GenePool^.Data.Ptr[Idx] := ThisGene^.Randomize; {Create a random genome}
    Scores := New_BigString(_Lng, BreedTo);
    Generations := 0;
  {Loop}
    repeat
      Inc(Generations);
    {Select the BreedFrom best unique genes}
      for Idx := 1 to BreedTo do {Score all genomes}
        Scores^.Data.Lng[Idx] := Score( PDR, Sizes,
                                        GenePtr(GenePool^.Data.Ptr[Idx]) );
      SortVector := Scores^.Sort;
      Breeders  := 0;
      GeneIndex := BreedTo; {The index of the best genome's index}
      repeat
        GeneNumber := SortVector^.Data.Wrd[GeneIndex]; {Get the gene's position}
        ThisGene   := GenePool^.Data.Ptr[GeneNumber];  {Get the gene}
        Duplicate  := False; {scan for dups}
        GeneNumber := 1;
        while (not Duplicate) and (GeneNumber <= Breeders) do
          begin
          Duplicate := ThisGene^.EqualTo(BreedingPool^.Data.Ptr[GeneNumber]);
          Inc(GeneNumber);
          end;
        if Duplicate
          then Dispose_BigString(ThisGene) {Duplicate of another top-N gene}
          else begin
               Inc(Breeders);
               BreedingPool^.Data.Ptr[Breeders] := ThisGene;
               end;
        Dec(GeneIndex);
      until (Breeders = BreedFrom) or (GeneIndex < 1);
      for GeneIndex := GeneIndex downto 1 do {zap the losers}
        Dispose_BigString(GenePool^.Data.Ptr[ SortVector^.Data.Wrd[GeneIndex] ]);
      BestGene  := BreedingPool^.Data.Ptr[1];
      BestScore := Scores^.Data.Lng[SortVector^.Data.Wrd[BreedTo]];
      if Report then
        begin
        MeanScore := 0;
        for Idx := 1 to BreedTo do Inc(MeanScore, Scores^.Data.Lng[Idx]);
        WriteLn(Output, 'Generation ', Generations,
                        ': Best score = ', BestScore: 7,
                        ' (', Floppies(PDR, Sizes, BestGene), ' floppies); ' +
                        'mean score = ', MeanScore div BreedTo: 7);
        end;
      Dispose_BigString(SortVector);
      AWinner := BestScore = PerfectScore;
    {Create BreedTo-BreedFrom mutated copies}
      TimeOut := Generations = MaxIterations;
      if not (AWinner or TimeOut) then
        begin {Create a new genepool}
        Move( BreedingPool^.Data.Ptr[1], GenePool^.Data.Ptr[1],
              BreedFrom * SizeOf(GenePtr)); {Move BreedingPool to GenePool}
        for Idx := BreedFrom + 1 to BreedTo do {create new genes}
          GenePool^.Data.Ptr[Idx] :=
            GenePtr( GenePool^.Data.Ptr[Random(BreedFrom)+1] )^.Mutate;
        end;
  {until Perfection or TimeOut}
    until AWinner or TimeOut;
    for Idx := 2 to BreedFrom do {Free remaining losers}
      Dispose_BigString( BreedingPool^.Data.Ptr[Idx] );
    Dispose_BigString(GenePool);
    Dispose_BigString(BreedingPool);
    Dispose_BigString(Scores);
    Optimize := BestGene;
end;


end.